
-- 
-- Fräse LS17
-- version: 1.0
-- date: 19.01.2017
-- author: chrisu70
-- 
-- Keine Veränderung ohne meine Erlaubnis!
-- No modification without	my permission!
--
-- History:
-- v1:	project started


Fraese = {};

function Fraese.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Overloading, specializations) and
	SpecializationUtil.hasSpecialization(Shovel, specializations) and 	
	SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations);
end;
function Fraese:load(savegame)
	self.getAllowFillShovel = Utils.overwrittenFunction(self.getAllowFillShovel, Fraese.getAllowFillShovel);
	self.ignoreVehicleDirectionOnLoad = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.fraese#ignoreVehicleDirectionOnLoad"), false);
	self.updatefraeseParticleSystems = Fraese.updatefraeseParticleSystems;
	self.setChangedFillLevelCallback = Fraese.setChangedFillLevelCallback;
	self.getCapacity = Utils.overwrittenFunction(self.getCapacity, Fraese.getCapacity);
	self.getFraeseEmptyingSpeed = Fraese.getFraeseEmptyingSpeed;
    self.fraeseTipReferenceNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.fraese#tipReferenceNode"));
    self.fraeseOpenNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.fraese#openNode"));
    self.fraeseOpenRotation = math.rad(Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#openRotation"), 0));
    self.fraeseOpenDirection = Utils.sign(Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.fraese#openDirection"), 1));
    self.fraeseEmptySpeed = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#emptySpeed"), self:getCapacity()/2)*0.001;
    self.fraeseFillLitersPerSecond = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#fillLitersPerSecond"), 10000) / 1000;
	
        
	self.effectsTimeout = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#effectsTimeout"), 0.1)*1000;
	self.fillEffectsTimeout = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#fillEffectsTimeout"), 0.1)*1000;
	self.fraeseFillEffects = EffectManager:loadEffect(self.xmlFile, "vehicle.fraeseFillEffect", self.components, self);
	
	self.isAlwaysTurnedOn = false;
	if self.shovelNodes ~= nil then
        self.isAlwaysTurnedOn = false;
    end
	self.fraese = {};
	self.fraese.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.fraese#fillUnitIndex"), 1);
    self.fraese.loadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.fraese#loadInfoIndex"), 1);
    self.fraese.node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.fraese#fraeseNode"));
    self.fraese.width = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#fraeseWidth"), 1.0);
    self.fraese.length = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#fraeseLength"), 0.5);
    self.fraese.yOffset = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.fraese#fraeseYOffset"), 0.0);
    self.fraese.fraeseRequiresMovement = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.fraese#fraeseRequiresMovement"), true);
    self.fraese.fraeseNeedsToBeTurnedOn = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.fraese#fraeseNeedsToBeTurnedOn"), false);
    self.fraese.fraeseNeedsActiveVehicle = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.fraese#fraeseNeedsActiveVehicle"), true);
	self.fraese.remainingDelta = 0;
	self.accumulatedFillDelta = 0;
    self.fillEffectActive = false;
    self.sentFillEffectActive = false;
    self.lastFillEffectActiveTime = 0;
	if self.fraeseDirtyFlag == nil then
        self.fraeseDirtyFlag = self:getNextDirtyFlag();
    end
end;

function Fraese:delete()
    if self.isClient then
        EffectManager:deleteEffects(self.fraeseFillEffects);
		
    end;
end
function Fraese:readUpdateStream(streamId, timestamp, connection)
	if connection.isServer then
        if streamReadBool(streamId) then
            local fillEffectActive = streamReadBool(streamId);
			self:updatefraeseParticleSystems(fillEffectActive);
        end
    end
end
function Fraese:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection.isServer then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.fraeseDirtyFlag) ~= 0) then
            streamWriteBool(streamId, self.sentFillEffectActive);
        end
    end
end
function Fraese:update(dt)
end 
function Fraese:updateTick(dt)
	if self.isServer then
		self.fillEffectActive = false;
		if self:getIsTurnedOn() then
			if (self:getIsActive() or not self.fraese.fraeseNeedsActiveVehicle) then
			local fraeseIsAllowed = true;
			local fraeseSpeed = 0;
			local fraese = self.fraese;
			if fraese.node ~= nil then
				if Vehicle.debugRendering then
					DebugUtil.drawDebugNode(fraese.node);
				end
				if self.fraeseRequiresMovement then
					if fraese.lastPos == nil then
						fraese.lastPos = {getWorldTranslation(fraese.node)};
					end
					local _,dy,dz = worldDirectionToLocal(getParent(fraese.node), 0,1,0);
					local alpha = math.atan2(dz, dy);
					setRotation(fraese.node, alpha, 0, 0);
	
					local x,y,z = worldToLocal(fraese.node, fraese.lastPos[1], fraese.lastPos[2], fraese.lastPos[3]);
					fraeseIsAllowed = z < -0.01;
					fraeseIsAllowed = fraeseIsAllowed and (alpha < math.pi/2) and (alpha > -math.pi/2);
	
					if fraeseIsAllowed then
						fraeseSpeed = math.min(1, Utils.vector3Length(x,y,z) / (dt*0.001));
					end
	
					fraese.lastPos[1], fraese.lastPos[2], fraese.lastPos[3] = getWorldTranslation(fraese.node);
				else
					fraeseSpeed = 1;
				end
	
				if self.fraese.fraeseNeedsToBeTurnedOn and self.getIsTurnedOn ~= nil and not self:getIsTurnedOn() then
					fraeseSpeed = 0;
				end
			end
			if fraeseIsAllowed then
				if self.fraeseOpenNode ~= nil then
					local x,_,_ = getRotation(self.fraeseOpenNode);
					if self.fraeseOpenDirection > 0 then
						if x < self.fraeseOpenRotation then
							fraeseIsAllowed = false;
						end;
					else
						if x > self.fraeseOpenRotation then
							fraeseIsAllowed = false;
						end
					end
				end
			end
			local emptySpeed = self:getFraeseEmptyingSpeed();
			local fillType = self:getUnitFillType(self.fraese.fillUnitIndex);
			local capacity = self.overloading.capacity;
			local fillLevel = self:getUnitFillLevel(self.overloading.fillUnitIndex);
			if self.fraese.node ~= nil and fraeseIsAllowed and fillLevel < capacity then
				local fraese = self.fraese;
	
				local length = fraese.length;
				local maxRadius = 1.75;
				local zPos = math.min(fraese.length/2, maxRadius);
	
				local x0,y0,z0 = localToWorld(fraese.node, -fraese.width, 0, zPos);
				local x1,y1,z1 = localToWorld(fraese.node, fraese.width, 0, zPos);
	
				local heapFillType = TipUtil.getFillTypeAtLine(x0,y0,z0, x1,y1,z1, zPos);
	
				if heapFillType ~= FillUtil.FILLTYPE_UNKNOWN and self:allowFillType(heapFillType) and self:allowUnitFillType(self.fraese.fillUnitIndex, heapFillType) then
	
					local emptyFactor = 1.0 - (emptySpeed / self.fraeseEmptySpeed);
					local fillFactor = fraeseSpeed * emptyFactor; -- * 0.001 * dt;
	
					local fraeseDelta = math.min(emptyFactor*(capacity - fillLevel), self.fraeseFillLitersPerSecond*dt + self.accumulatedFillDelta);
	
					fraeseDelta = fraeseDelta * math.min(1, fillFactor);
	
					if length > maxRadius * 2 then
						fraeseDelta = fraeseDelta / math.floor(self.fraese.length / (maxRadius * 2));
					end;
	
					while true do
						local sx,sy,sz = localToWorld(fraese.node, -fraese.width, fraese.yOffset, zPos);
						local ex,ey,ez = localToWorld(fraese.node,  fraese.width, fraese.yOffset, zPos);
	
						local sh = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, sx,sy,sz);
						local eh = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, ex,ey,ez);
						if sh < 0.1 then
							sy = sh;
						end
						if eh < 0.1 then
							ey = eh;
						end
	
						local innerRadius = math.min(length, maxRadius); --0.05;
						local radius = nil; --1.0;
	
						fraese.remainingDelta = fraese.remainingDelta + fraeseDelta;
						if fraese.remainingDelta > 20 then
							local fillLevel = self:getUnitFillLevel(self.fraese.fillUnitIndex);
							local capacity = self.overloading.capacity;
							fraeseDelta = math.min(fraese.remainingDelta, capacity-fillLevel);
							fraese.remainingDelta = fraese.remainingDelta - fraeseDelta;
						else
							fraeseDelta = 0;
						end;
						local deltaFill, lineOffset = TipUtil.tipToGroundAroundLine(self, -fraeseDelta, heapFillType, sx,sy,sz, ex,ey,ez, innerRadius, radius, self.fraese.lineOffsetfraese, true, nil);
						self.fraese.lineOffsetfraese = lineOffset;
	
						self.accumulatedFillDelta = math.min(fraeseDelta + deltaFill, TipUtil.getMinValidLiterValue(heapFillType));
	
						if deltaFill < 0 then
							self:setUnitFillLevel(self.fraese.fillUnitIndex, fillLevel - deltaFill, heapFillType, false, self.fillVolumeLoadInfos[self.fraese.loadInfoIndex]);
							self.fillEffectActive = true;
						elseif deltaFill ~= nil then
							self.fillEffectActive = false;
						if self.changedFillLevelCallback ~= nil then
								if self.changedFillLevelCallbackTarget ~= nil then
									self.changedFillLevelCallback(self.changedFillLevelCallbackTarget, self, deltaFill, heapFillType);
								else
									self.changedFillLevelCallback(self, deltaFill, heapFillType);
								end
							end
						end
					if length > maxRadius * 2 then
							length = length - maxRadius * 2;
							if length > maxRadius * 2 then
								zPos = zPos + maxRadius * 2;
							else
								zPos = zPos + (fraese.length-zPos+maxRadius)/2;
								length = fraese.length-zPos;
							end;
						else
							break;
						end;
					end;
				else
					self.accumulatedFillDelta = 0;
					self.fillEffectActive = false;
					
				end
	
			end
		end
	end
	if self.fillEffectActive ~= self.sentFillEffectActive then
			self:raiseDirtyFlags(self.fraeseDirtyFlag);
			self.sentFillEffectActive = self.fillEffectActive;
			self:updatefraeseParticleSystems(self.fillEffectActive);
		end;
	
	end
    if self.isClient then

        if not self.fillEffectActive  then
            self.lastFillEffectActiveTime = 0;
            if self.lastFillPS ~= nil then
                ParticleUtil.setEmittingState(self.lastFillPS, false);
            end
			if self.fraeseFillEffects ~= nil then
                EffectManager:stopEffects(self.fraeseFillEffects);
            end
        end
    end;
end
function Fraese:getAllowFillFraese(fillType)
    if self.shovelOpenNode ~= nil then
        local x,_,_ = getRotation(self.shovelOpenNode);
        if self.shovelOpenDirection > 0 then
            if x < self.shovelOpenRotation then
                return false;
            end;
        else
            if x > self.shovelOpenRotation then
                return false;
            end;
        end
    end;
    if self:getShovelEmptyingSpeed() > 0 then
        return false;
    end;
    return self:allowFillType(fillType, false)
end;
function Fraese:updatefraeseParticleSystems(fillEffectActive)
    if self.isClient then
        local fillType = self:getUnitFillType(self.fraese.fillUnitIndex);
        if fillEffectActive then
            self.lastFillEffectActiveTime = g_currentMission.time;
            if self.fraeseFillEffects ~= nil then
                EffectManager:setFillType(self.fraeseFillEffects, fillType)
                EffectManager:startEffects(self.fraeseFillEffects);
            end;
		end
    end;
end
function Fraese:getAllowFillShovel(superFunc, fillType)
 if self:getIsTurnedOn() then
	 if superFunc ~= nil then
		 return superFunc(self, fillType);
	 end
	
 end
 return false;
end;
function Fraese:setUnitFillLevel(superFunc, fillUnitIndex, fillLevel, fillType, force, fillInfo)
    local fillLevel = self.fillUnits[self.fraese.fillUnitIndex].fillLevel;

    superFunc(self, fillUnitIndex, fillLevel, fillType, force, fillInfo);
	
end
function Fraese:setChangedFillLevelCallback(callback, callbackTarget)
    self.changedFillLevelCallback = callback;
    self.changedFillLevelCallbackTarget = callbackTarget;
end
function Fraese:getFraeseEmptyingSpeed()
    if self.fraeseOpenNode ~= nil then
        local x,_,_ = getRotation(self.fraeseOpenNode);
        if self.fraeseOpenDirection > 0 then
            if x < self.fraeseOpenRotation then
                return 0;
            end;
        else
            if x > self.fraeseOpenRotation then
                return 0;
            end;
        end
    end;
    return 0;
end;
function Fraese:draw()
end;
function Fraese:getCapacity(superFunc)
return 0;
end;
function Fraese:mouseEvent(posX, posY, isDown, isUp, button)
end;
function Fraese:keyEvent(unicode, sym, modifier, isDown)
end;